(*****************************************************************************)
(* The analysis itself: generation function, coding of non-empty             *)
(* intersection, and implementation of the attacker                          *)
(*                                                                           *)
(*                                 Time-stamp: <14-01-2004 Mikael Buchholtz> *)
(*****************************************************************************)
(* Assumes that all names, variables, crypto-points, and labels only contain *)
(* alphanumeric characters, prime (') or underscore (_) and should not begin *)
(* with an _. If not, the ALFP formula generated cannot be parsed by the     *)
(* Succinct Solver.                                                          *)
(*****************************************************************************)

structure Analysis1 =
struct
local

    open LLysa

    exception Problem of string

    (*************************************************************************)
    (*                                                                       *)
    (* Auxiliary functions to generate bits and pieces of an ALPF formula.   *)
    (*                                                                       *)
    (*************************************************************************)

    fun after []      e = []
      | after (l::ls) e = l::e::after ls e


    fun between []      e = []
      | between [t]     e = [t]
      | between (l::ls) e = l::e::between ls e


    fun quantify(l,u,Q,x) =
	let 
	    val s =  Q^" "^x^"_"^(Int.toString l)^". "  
	in
	    if l > u then ""
	    else s^quantify(l+1,u,Q,x)
	end


    fun queryVar(l,u,x) =
	let 
	    val s = x^"_"^(Int.toString l)
	in
	    if l > u then "" 
	    else if l = u then s
	    else s^", "^queryVar(l+1,u,x)
	end


    fun queryRel(l,u,R,x) =
	let 
	    val s = R^"("^x^"_"^(Int.toString l)^")"
	in
	    if l > u then "" 
	    else if l = u then s
	    else s^" & "^queryRel(l+1,u,R,x)
	end


    fun queryRels(l,u,R,x1,x2) =
	let 
	    val s = R^"("^x1^"_"^(Int.toString l)^", "^x2^"_"^(Int.toString l)^")"
	in
	    if l > u then "" 
	    else if l = u then s
	    else s^" & "^queryRels(l+1,u,R,x1,x2)
	end
    
    fun getLab(NAME(_,l))         = l
      | getLab(NAMEP(_,l))        = l
      | getLab(NAMEM(_,l))        = l
      | getLab(VAR(_,l))          = l
      | getLab(ENC(_,_,_,_,_,l))  = l
      | getLab(AENC(_,_,_,_,_,l)) = l


    fun labList el =
	String.concat(between (List.map getLab  el) ", ")
    
    fun labListAfter el =
	String.concat(after (List.map getLab  el) ", ")
    
		  

    fun NEIList(el, v, i) =
	let fun NEIify([]   ,v,i) = ["1"]
	      | NEIify(e::el,v,i) = 
		let val n = "NEI("^v^"_"^(Int.toString(i))^", "^(getLab(e))^")"
		in n::NEIify(el,v,i+1)
		end
	in
	    String.concat(between (NEIify(el,v,i)) " & ")
	end

(*
    fun MSList(el, v, i) =
	let fun MSify([]   ,v,i) = ["1"]
	      | MSify(e::el,v,i) = 
		let val n = "MS("^v^"_"^(Int.toString(i))^", "^(getLab(e))^")"
		in n::MSify(el,v,i+1)
		end
	in
	    String.concat(between (MSify(el,v,i)) " & ")
	end
*)

    fun cryptoList(cl, v) =
	String.concat(after (List.map (fn c => "D("^v^", "^c^")")  cl) " & ")


    fun copyC(l,u,l1,l2) =
	let val s = "(A U. C("^l1^"_"^Int.toString(l)^", U) =>"^" C("^l2^", U))"
	in
	    if u = 0 then " 1 "
	    else if l>u then ""
	    else if l=u then s
	    else s^" & "^copyC(l+1,u,l1,l2)
	end

    fun querySameVar(0,v) = ""
      | querySameVar(1,v) = v
      | querySameVar(k,v) = v^", "^querySameVar(k-1,v)


(*
    fun toStringSymbol(s) =
	let fun disallowedChar c = not (Char.isAlphaNum(c) orelse
					c = #"_" orelse
					c = #"'")
	in
	    if List.exists disallowedChar (String.explode s)
	    then raise Problem ("Symbol '"^s^"' contains disallowed character")
	    else s
	end			 
*)
in 

fun canon(NAME(n,_))  = n
  | canon(NAMEP(n,_)) = "_PLUS_"^n
  | canon(NAMEM(n,_)) = "_MINUS_"^n
  | canon(VAR(x,_))   = x
  | canon(_)          = 
    raise Problem "canon only defined for names and variables"


local 
    fun insertRho(xl, v, i) =
	let fun rhoify([]   ,v,i) = []
	      | rhoify(x::xl,v,i) = 
		let val r = "RHO("^canon(x)^", "^v^"_"^Int.toString(i)^")" 
		in r::rhoify(xl,v,i+1) 
		end	
	in     
	    String.concat(after (rhoify(xl, v, i)) " & ")
	end
in


(*****************************************************************************)
(*                                                                           *)
(* genTerm: LLysa.Term -> ALFP logic                                         *)
(* genProc: LLysa.Proc -> ALFP logic                                         *)
(*                                                                           *)
(* Corresponds to the function {\cal G} in the paper                         *)
(* M. Buchholtz: "Implementing Control Flow Analysis for Security Protocols" *)
(*                                                                           *)
(*****************************************************************************)

fun genTerm(NAME(n,l))  = "C("^l^", "^canon(NAME(n,""))^")\n"
			  
  | genTerm(NAMEP(n,l)) = "C("^l^", "^canon(NAMEP(n,""))^") & "^
			  "KP("^canon(NAMEP(n,""))^", "^canon(NAMEM(n,""))^") & "^
			  "KP("^canon(NAMEM(n,""))^", "^canon(NAMEP(n,""))^")\n"

  | genTerm(NAMEM(n,l)) = "C("^l^", "^canon(NAMEM(n,""))^") & "^
			  "KP("^canon(NAMEP(n,""))^", "^canon(NAMEM(n,""))^") & "^
			  "KP("^canon(NAMEM(n,""))^", "^canon(NAMEP(n,""))^")\n"
			  
  | genTerm(VAR(x,l))   = "(A L'. RHO("^canon(VAR(x,""))^", L') => "^
			  "(A U.  C(L', U) => C("^l^", U)))\n"

  | genTerm(ENC(el,e0,c,cl,ls,l)) 
                        = genTerm(e0)^" & "^
			  genTermList(el)^" & \n"^				      
			  "SIGMA"^Int.toString(List.length(el))^
			  "("^l^", "^getLab(e0)^", "^
                          (labListAfter el)^c^", "^ls^") & \n"^
			  cryptoList(cl,ls)^
			  "C("^l^", "^l^")"
			  
			  
  | genTerm(AENC(el,e0,c,cl,ls,l)) =                           
                          genTermList((e0::el))^" &\n"^
			  "ALPHA"^Int.toString(List.length(el))^
			  "("^l^", "^getLab(e0)^", "^
			  (labListAfter el)^c^", "^ls^") & \n"^
			  cryptoList(cl,ls)^
			  "C("^l^", "^l^")"
			  


and genTermList el        = String.concat(between ("1"::(List.map genTerm el)) " & " )


fun genProc(OUT(el,p))    = let val k = List.length el 
			    in
				"\n/* OUTPUT - length "^Int.toString(k)^" */\n"^
				genTermList el ^ " & \n" ^ 
				"KAPPA"^(Int.toString(k))^"("^labList el^") & \n" ^
				"("^genProc(p)^")"
			    end

  | genProc(INP(el,xl,p)) = let val j  = List.length el 
				val k  = j + List.length xl 
				val lp = "L'"
			    in 
				"\n/* INPUT - length "^Int.toString(k)^" */"^
				genTermList el ^ " & \n" ^
				"("^quantify(1,k,"A",lp)^"\n"^
				"KAPPA"^(Int.toString(k))^"("^queryVar(1,k,lp)^") =>\n"^
				"(("^NEIList(el,lp,1)^") => \n" ^ 
				insertRho(xl,lp,j+1)^"\n"^
				"("^genProc(p)^")))"
			    end
  
  | genProc(DEC(e,el,xl,e0,c,cl,ls,p)) = 
    let val j  = (List.length el)
	val k  = (List.length xl)+j
	val lp = "L'"
    in
	"\n/* DECRYPTION - length "^Int.toString(k)^" */\n"^
	genTerm(e)^" & \n"^
	cryptoList(cl,ls)^"\n"^
	genTerm(e0)^" & "^
	genTermList(el)^" & \n"^
	"(A L. "^quantify(0,k,"A",lp)^"A ELL0. A L0S. A L'. \n"^
	"SIGMA"^Int.toString(k)^"(L',"^queryVar(0,k,lp)^", ELL0, L0S) & "^
	"C("^getLab(e)^", L') => \n"^
	"(("^NEIList((e0::el),lp,0)^") => \n" ^ 
	insertRho(xl,lp,j+1)^
	"PQD(L0S, "^c^", "^ls^", ELL0) & \n"^
	" ("^genProc(p)^")))"
    end


  | genProc(ADEC(e,el,xl,e0,c,cl,ls,p)) = 
    let val j  = (List.length el)
	val k  = (List.length xl)+j
	val lp = "L'"
    in
	"\n/* ASYMMETRIC DECRYPTION - length "^Int.toString(k)^" */\n"^
	genTerm(e)^" &\n"^
	cryptoList(cl,ls)^"\n"^
	genTermList(e0::el)^" &\n"^
	"(A "^lp^". "^quantify(0,k,"A",lp)^" A ELL'. A LS'. \n"^
	"ALPHA"^Int.toString(k)^"("^lp^", "^queryVar(0,k,lp)^", ELL', LS') & "^
	"C("^getLab(e)^", "^lp^") =>\n"^
	"(E V0. E V0'. C("^getLab(e0)^", V0) & C("^lp^"_0, V0') & KP(V0,V0') &"^
	NEIList(el,lp,1)^") =>\n"^
	insertRho(xl,lp,j+1)^
	"PQD(LS', "^c^", "^ls^", ELL') & \n"^
	" ("^genProc(p)^"))"
    end


(*****************************************************************************)
(* Below is an analysis with an explicit reachability component rather than  *)
(* just using =>. Works more efficently when input/decryption matches lots   *)
(* of values. The analysis below still needs some work before it can be used *)
(* for real.                                                mib - 26.09.2003 *)
(*****************************************************************************)
(*
  | genProc(INP(el,xl,p)) = let val j  = List.length el 
				val k  = j + List.length xl 
				val lp = "L'"
			    in 
				"\n/* INPUT - length "^Int.toString(k)^" */"^
				genTermList el ^ " & \n" ^
				"("^quantify(1,k,"A",lp)^"\n"^
				"KAPPA"^(Int.toString(k))^"("^queryVar(1,k,lp)^") =>\n"^
				"(("^NEIList(el,lp,1)^") => \n" ^ 
				insertRho(xl,lp,j+1)^
				"REACH("^getLab(List.hd(xl))^"))) & \n"^
				"(REACH("^getLab(List.hd(xl))^") => "^genProc(p)^")"
				(* FIXME: reachability does not work for input when xl is empty *)
			    end
  
  | genProc(DEC(e,el,xl,e0,c,cl,ls,p)) = 
    let val j  = (List.length el)
	val k  = (List.length xl)+j
	val lp = "L'"
    in
	"\n/* DECRYPTION - length "^Int.toString(k)^" */\n"^
	genTerm(e)^" & \n"^
	cryptoList(cl,ls)^" & \n"^
	genTerm(e0)^" & "^
	genTermList(el)^" & \n"^
	"(A L. "^quantify(0,k,"A",lp)^"A ELL0. A L0S. A L'. \n"^
	"SIGMA"^Int.toString(k)^"(L',"^queryVar(0,k,lp)^", ELL0, L0S) & "^
	"C("^getLab(e)^", L') => \n"^
	"(("^NEIList((e0::el),lp,0)^") => \n" ^ 
	insertRho(xl,lp,j+1)^
	"PQD(L0S, "^c^", "^ls^", ELL0) & \n"^
	"REACH("^ls^"))) &\n"^
	"(REACH("^ls^") => "^genProc(p)^")"
	
    end


  | genProc(ADEC(e,el,xl,e0,c,cl,ls,p)) = 
    let val j  = (List.length el)
	val k  = (List.length xl)+j
	val lp = "L'"
    in
	"\n/* ASYMMETRIC DECRYPTION - length "^Int.toString(k)^" */\n"^
	genTerm(e)^" &\n"^
	cryptoList(cl,ls)^" &\n"^
	genTermList(e0::el)^" &\n"^
	"(A "^lp^". "^quantify(0,k,"A",lp)^" A ELL'. A LS'. \n"^
	"ALPHA"^Int.toString(k)^"("^lp^", "^queryVar(0,k,lp)^", ELL', LS') & "^
	"C("^getLab(e)^", "^lp^") =>\n"^
	"(E V0. E V0'. C("^getLab(e0)^", V0) & C("^lp^"_0, V0') & KP(V0,V0') &"^
	NEIList(el,lp,1)^") =>\n"^
	insertRho(xl,lp,j+1)^
	"PQD(LS', "^c^", "^ls^", ELL') & \n"^
	"REACH("^ls^")) &\n"^
	"(REACH("^ls^") => "^genProc(p)^")"
	
	
    end
*)
  | genProc(NEW(n,p))     = "\n/* RESTRICTION */\n"^"("^genProc(p)^")"

  | genProc(ANEW(n,p))    = "\n/* ASYMMETRIC RESTRICTION */\n"^"("^genProc(p)^")"

  | genProc(BANG(p))      = "\n/* REPLICATION */\n"^"("^genProc(p)^")"

  | genProc(PAR(p1,p2))   = "\n/* PARALLEL */\n"^"("^genProc(p1)^") & ("^genProc(p2)^")"

  | genProc(NIL)          = "\n/* NIL */\n"^"1"



(*****************************************************************************)
(* Values used by the attacker                                               *)
(*****************************************************************************)

val LDY  = "_L_DY"
val LSDY = "_L_C"

val NDY  = NAME("_NDY",LDY)
val NPDY = NAMEP("_NDY",LDY)
val NMDY = NAMEM("_NDY",LDY)
val ZDY  = VAR("_ZDY",LDY)

(*****************************************************************************)
(*                                                                           *)
(* Functions to generate auxiliary components for the analysis               *)
(*                                                                           *)
(* genNEI : Proc -> ALFP logic                                               *)
(*       Generates the predicate NEI of all pairs of labels, which represent *)
(*       languages with non-empty intersections.                             *)
(*                                                                           *)
(* includes the function                                                     *)
(*                                                                           *)
(* genNames : Proc -> ALFP logic                                             *)
(*       Generates the predicate N containing all names                      *)
(*                                                                           *)
(*****************************************************************************)

fun genNEI p =
    let val eal  = sencArityProc(p)
	val aeal = aencArityProc(p)
		   
	fun genNames(p) = String.concat (between 
					 (List.map 
					      (fn n => "N("^canon(n)^")") 
					      (NDY::NPDY::NMDY::
					       namesProc(p))) " & ")

	fun genEnc k = 
	    "\n(A L1. A L2.  \n\t("^quantify(0,k,"E","L1")^"E ELL1. E LS1. E L1'.\n\t "^
 	    quantify(0,k,"E","L2")^"E ELL2. E LS2. E L2'.\n\t\t"^		       
	    "SIGMA"^(Int.toString(k))^"(L1',"^queryVar(0,k,"L1")^", ELL1, LS1) &\n\t\t" ^ 
	    "SIGMA"^(Int.toString(k))^"(L2',"^queryVar(0,k,"L2")^", ELL2, LS2) &\n\t\t"^
	    queryRels(0,k,"NEI","L1","L2")^"&\n\t\t"^ 
	    "C(L1, L1') &\n\t\t"^		       
	    "C(L2, L2')"^		       
	    ") =>\n\t\t\tNEI(L1,L2))\n"

	fun genAenc k = 
	    "\n(A L1. A L2.  \n\t("^quantify(0,k,"E","L1")^"E ELL1. E LS1. E L1'.\n\t "^
 	    quantify(0,k,"E","L2")^"E ELL2. E LS2. E L2'.\n\t\t"^		       
	    "ALPHA"^(Int.toString(k))^"(L1',"^queryVar(0,k,"L1")^", ELL1, LS1) &\n\t\t" ^ 
	    "ALPHA"^(Int.toString(k))^"(L2',"^queryVar(0,k,"L2")^", ELL2, LS2) &\n\t\t"^
	    queryRels(0,k,"NEI","L1","L2")^"&\n\t\t"^ 
	    "C(L1, L1') &\n\t\t"^		       
	    "C(L2, L2')"^		       
	    ") =>\n\t\t\tNEI(L1,L2))\n"
     in
	genNames(p)^" & \n\n"^

	"(A L1. A L2.\n\t(E U. N(U) & C(L1,U)  & C(L2,U))\n\t\t=> NEI(L1,L2))"^
	(String.concat(map (fn x => "&\n"^genEnc(x)) eal))^
	(String.concat(map (fn x => "&\n"^genAenc(x)) aeal))
    end



(*****************************************************************************)
(* genDY generates the attacker via the constructive approach suggested by   *)
(* the proof of Theorem 4                                                    *)
(*****************************************************************************)

fun genDY(p) =
let
    
    val eal  = sencArityProc(p)
    val aeal = aencArityProc(p)
    val kal  = Set.union([1],kappaArityProc(p))
    val nf   = fnProc(p)			    
    val C    = Set.union([CPDY], cryptoProc(p))
			    
	
    fun parList([])    = NIL
      | parList([p])   = p
      | parList(p::pl) = List.foldl (fn(p1,p2) => PAR(p1,p2)) p pl

    fun zdyList(0) = []
      | zdyList(k) = ZDY::zdyList(k-1)

    fun inpSeq(0, p) = p
      | inpSeq(k, p) = INP([],[ZDY],inpSeq((k-1),p))

    fun R1(k) = INP([],zdyList(k),NIL)
    
    fun R2(k) = INP([],[ZDY],
		INP([],[ZDY],
		DEC(ZDY,[],zdyList(k),ZDY,CPDY,C,LSDY,NIL)))

    fun R3(k) = inpSeq(k,OUT([ENC(zdyList(k),ZDY,CPDY,C,LSDY,LDY)],NIL))

    fun R4(k) = inpSeq(k,OUT(zdyList(k),NIL))

    val R5    = PAR(parList
			(List.map (fn NAME(n,_)  => OUT([NAME(n,LDY)], NIL)
				     | _         => NIL ) 
				   (NDY::fnProc(p))),
		     INP([],[ZDY],NIL))

    fun R6(k) = INP([],[ZDY],
		INP([],[ZDY],
                ADEC(ZDY,[],zdyList(k),ZDY,CPDY,C,LSDY,NIL)))

    fun R7(k) = inpSeq(k,OUT([AENC(zdyList(k),ZDY,CPDY,C,LSDY,LDY)],NIL))

    val R8    = PAR(parList
		    (List.map (fn NAMEP(n,_) => OUT([NAMEP(n,LDY)],NIL)
				| NAMEM(n,_) => OUT([NAMEM(n,LDY)],NIL)
				| _          => NIL)
			      (NPDY::NMDY::fnProc(p))),
		    INP([],[ZDY],NIL))
		 
in
     genProc(parList(
		     (List.map R1 kal)@
		     (List.map R2 eal)@
		     (List.map R3 eal)@
		     (List.map R4 kal)@ 
		     [R5]@
		     (List.map R6 aeal)@
		     (List.map R7 aeal)@
		     [R8]))
end


(*****************************************************************************)
(* genDY2 an optimised version of the attacker genDY2, where ZDY is only     *)
(* bound once... Should be OK with current analysis, but may not be so for   *)
(* more sophisitcated analyses.                                              *)
(*****************************************************************************)

fun genDY2(p) =
let
    
    val eal = sencArityProc(p)
    val aeal= aencArityProc(p)
    val kal = Set.union([1],kappaArityProc(p))
    val nf  = fnProc(p)			    
    val C   = Set.union([CPDY], cryptoProc(p))
			    	
    fun parList([])    = NIL
      | parList([p])   = p
      | parList(p::pl) = List.foldl (fn(p1,p2) => PAR(p1,p2)) p pl

    fun zdyList(0) = []
      | zdyList(k) = ZDY::zdyList(k-1)

    fun ldyName(NAME(n,_))  = NAME(n,LDY)
      | ldyName(NAMEP(n,_)) = NAMEP(n,LDY)
      | ldyName(NAMEM(n,_)) = NAMEM(n,LDY)
      | ldyName(_)          = raise Problem "ldyName only defined for names"
 

    fun R1(k) = INP([],zdyList(k),NIL)
    
    fun R2(k) = DEC(ZDY,[],zdyList(k),ZDY,CPDY,C,LSDY,NIL)

    fun R3(k) = OUT([ENC(zdyList(k),ZDY,CPDY,C,LSDY,LDY)],NIL)

    fun R4(k) = OUT(zdyList(k),NIL)

    fun R6(k) = ADEC(ZDY,[],zdyList(k),ZDY,CPDY,C,LSDY,NIL)
		
    fun R7(k) = OUT([AENC(zdyList(k),ZDY,CPDY,C,LSDY,LDY)],NIL)

    val R5    = parList
	        (List.map (fn n => OUT([ldyName(n)], NIL)) (NDY::fnProc(p)))

    val R8    = OUT([NPDY],OUT([NMDY],NIL))
in
     genProc( 
	    (parList(
		     (List.map R1 kal)@
		     (List.map R2 eal)@
		     (List.map R3 eal)@
		     (List.map R4 kal)@ 
		     [R5]@
		     (List.map R6 aeal)@ 
		     (List.map R7 aeal)@ 
		     [R8])))
end


(*****************************************************************************)
(* genDYCSFW - generates Dolve Yao condition -- now obsolete                 *)
(* used for the experiments published in CSFW'03                             *)
(*****************************************************************************)
(* Not updated to handle asymmetric cryptography!!!                          *)
(*****************************************************************************)
(*
fun genDYCSFW(p) =
    let
	val eal = sencArityProc(p)
	val kal = kappaArityProc(p)
		  

	fun genDY1(k) =	"("^quantify(1,k,"A","L")^
			"KAPPA"^Int.toString(k)^"("^
			queryVar(1,k,"L")^") => "^
			"("^copyC(1,k,"L",LDY)^"))"
   
	fun genDY2(k) = "(A L. "^quantify(0,k,"A","L")^"A ELL. A LS."^
			"SIGMA"^Int.toString(k)^"(L,"^queryVar(0,k,"L")^", ELL, LS) & "^
			"C("^LDY^", L) => "^
			"(NEI(L_0, "^LDY^") => "^
			"(("^copyC(1,k,"L",LDY)^") & "^
			"(PQD(LS, "^CPDY^", "^LSDY^", ELL)))))" 

	fun genDY3(k) =	"SIGMA"^Int.toString(k)^
			"("^querySameVar(k+2,LDY)^", "^CPDY^", "^LSDY^") & "^
			"C("^LDY^", "^LDY^")"
		
	fun genDY4(k) = "KAPPA"^Int.toString(k)^"("^
			querySameVar(k,LDY)^")"
			
	val genDY5    = String.concat(List.map  
					  (fn n => "C("^LDY^", "^canon(n)^") &\n")
					  (fnProc(p)))^
			"C("^LDY^", "^canon(NDY)^")\n"
		
    in	
	(String.concat (after (map genDY1 kal) " & "))^"\n"^
	(String.concat (after (map genDY2 eal) " & "))^"\n"^ 
	(String.concat (after (map genDY3 eal) " & "))^"\n"^
	cryptoList(CPDY::(cryptoProc(p)), LSDY)^" & \n"^
	(String.concat (after (map genDY4 kal) " & "))^"\n"^
	genDY5
	
    end
*)


(*****************************************************************************)
(*                                                                           *)
(* analyse: Lysa.Proc -> ALFP logic                                         *)
(*                                                                           *)
(* Main fuction: generates the formula for the analysis of a process         *)
(* including DY attacker and auxiliary predicates N and NEI.                 *)
(*                                                                           *)
(*                                                                           *)
(*****************************************************************************)

fun analyse(p) =    

    "("^
    "/****************************************************************/\n"^
    "/* Clauses for the process                                      */\n"^
    "/****************************************************************/\n\n"^

    "(\n"^genProc(p)^") & \n"^

    "/****************************************************************/\n"^
    "/* Non-empty intersection                                       */\n"^
    "/****************************************************************/\n\n"^

    "(\n"^genNEI(p)^") & \n"^

    "/****************************************************************/\n"^
    "/* Dolve-Yao formula                                            */\n"^
    "/****************************************************************/\n\n"^

    "(\n"^genDY2(p)^")) & \n"^

    "/****************************************************************/\n"^
    "/* Pending Queries to PSI                                       */\n"^
    "/****************************************************************/\n\n"^

    "(A L0S. A ELL. A LS. A ELL0. PQD(L0S, ELL, LS, ELL0) => "^
    "((!D(L0S, ELL) | !D(LS, ELL0)) => PSI(ELL0, ELL)))"



(*****************************************************************************)
(* Auxiliary relations for observation purposes are given below to           *)
(* "Experiments 1" given below. Not really part of the analysis!             *)
(*****************************************************************************)

fun genObs(p,oul) =
    let
	val kal = LLysa.kappaArityProc(p)
	fun queryNotLdy(l,u,v) =
	    let val s = v^"_"^Int.toString(l)^" != "^LDY
	    in
		if u = 0 then " 1 "
		else if l>u then ""
		else if l=u then s
		else s^" & "^queryNotLdy(l+1,u,v)
	    end

	fun genVarObs(VAR(x,_)) = "(A u. UF_RHO("^x^", u) => OBS_VAR("^x^", u))"
	  | genVarObs(_)         = raise Problem "OBS_VAR is suppose to be a variable"

    in
	"(A U. C("^LDY^",U) => DY(U)) & "^	
	"(A X. A L. A U. RHO(X, L) & L != "^LDY^" & C(L, U) => UF_RHO(X,U)) &  "^
	"(A X. RHO(X, "^LDY^") => UF_RHO(X, ALL_FROM_DY)) & "^
	"(A U. C("^LDY^",U) & N(U) => DYN(U)) & "^
	(String.concat(between ("1"::(List.map genVarObs oul)) " & "))
    end
end
end

end;