(*****************************************************************************)
(* LysaASCIIIO                                                               *)
(*                                                                           *)
(* A structure for parsing and outputting LySa to and from ASCII format      *)
(*                                                                           *)
(*                                 Time-stamp: <14-01-2004 Mikael Buchholtz> *)
(*****************************************************************************)

structure LysaASCIIIO =
struct

local

    (*************************************************************************)
    (* Declarations for parsing                                              *)
    (*************************************************************************)

    open Lysa

    structure LysaLrVals = 
    LysaLrValsFun(structure Token = LrParser.Token)

    structure LysaLexer =
    LysaLexerFun(structure Tokens = LysaLrVals.Tokens)

    structure LysaParser = Join(structure ParserData = LysaLrVals.ParserData
				structure Lex        = LysaLexer
				structure LrParser   = LrParser)


    fun toVar gamma (NAME(n)) = if Set.isMember(VAR(n),gamma) 
				then VAR(n) else NAME(n)
      | toVar gamma (ENC(ts,t,c,a)) 
                              = ENC(List.map (toVar gamma) ts,(toVar gamma) t,c,a)
      | toVar gamma (AENC(ts,t,c,a)) 
                              = AENC(List.map (toVar gamma) ts,(toVar gamma) t,c,a)
      | toVar gamma (t)       = t


    fun bindVar gamma (OUT(tl,P))      = OUT(List.map (toVar gamma) tl,bindVar gamma P)
      | bindVar gamma (INP(ts,xs,P))   = INP(List.map (toVar gamma) ts,
					     xs,bindVar (Set.union(gamma,xs)) P)
      | bindVar gamma (DEC(t,ts,xs,tk,c,F,P)) 
                                       = DEC((toVar gamma) t,List.map (toVar gamma) ts, xs,
					     (toVar gamma) tk,c,F,bindVar (Set.union(gamma,xs)) P)
      | bindVar gamma (ADEC(t,ts,xs,tk,c,F,P)) 
                                       = ADEC((toVar gamma) t,List.map (toVar gamma) ts,xs,
					      (toVar gamma) tk,c,F,bindVar (Set.union(gamma,xs)) P)
                                  
      | bindVar gamma (NEW(n,P))       = NEW(n,bindVar gamma P)
      | bindVar gamma (ANEW(n,P))      = ANEW(n,bindVar gamma P)
      | bindVar gamma (BANG P)         = BANG(bindVar gamma P)
      | bindVar gamma (PAR Ps)         = PAR(List.map (bindVar gamma) Ps)
      | bindVar gamma (NIL)            = NIL
      | bindVar gamma (PAR_I(i,a,P))   = PAR_I(i,a,bindVar gamma P)
      | bindVar gamma (PAR_X(i,a,j,P)) = PAR_X(i,a,j,bindVar gamma P)
      | bindVar gamma (NEW_I(il,P))    = NEW_I(il,bindVar gamma P)
      | bindVar gamma (ANEW_I(il,P))   = ANEW_I(il,bindVar gamma P)
      | bindVar gamma (_)              = raise Lysa.Problem 
						   "This construct cannot be produced by the parser"
in

exception parseError of string

fun parseStream stream =
    let val lexer    = LysaParser.makeLexer (fn _ => TextIO.inputLine stream)
	val dummyEOF = LysaLrVals.Tokens.EOF(0,0)
		       
	fun errorMessage(s,i,j) = 
	    raise parseError ("line "^(Int.toString i)^": "^s)
	
	val (result, lexer)    = LysaParser.parse(0,lexer, errorMessage,())
	val (nextToken, lexer) = LysaParser.Stream.get lexer
    in
	if LysaParser.sameToken(nextToken,dummyEOF) 
	then bindVar [] result
	else raise parseError "parsing ended before end-of-file"		   
    end
	handle LysaLexer.UserDeclarations.lexError s => raise parseError s
end

fun parseFile filename = parseStream (TextIO.openIn filename)

fun parseString str    = parseStream (TextIO.openString str)


local

    (*************************************************************************)
    (* Declarations for ASCII output                                         *)
    (*************************************************************************)
    
    open Lysa

    fun between []      e = ""
      | between [t]     e = t
      | between (l::ls) e = l^e^(between ls e)

    fun toStringNum (n)              = if n >= 0 
				       then Int.toString(n) 
				       else "-"^Int.toString(~1*n)

    fun toStringSymb(n,[])           = if (n,[]) = Lysa.CPDY
				       then "CPDY" else n
      | toStringSymb(n,tl)           = if (n,tl) = Lysa.CPDY
				       then "CPDY" else n^"_{"^(between tl ",")^"}"
	
    fun toStringSymbPM s (n,[])      = n^s
      | toStringSymbPM s (n,tl)      = n^s^"_{"^(between tl ",")^"}"
				     
    fun toStringSymbList([])         = ""
      | toStringSymbList([s])        = toStringSymb(s)
      | toStringSymbList(s::sl)      = toStringSymb(s)^","^toStringSymbList(sl)


    fun toStringDest(SOME(DEST(cl))) = " dest {"^toStringSymbList(cl)^"}"
      | toStringDest(NONE)           = ""
      | toStringDest(_)              = raise Problem "unsupported annotation"	     
	
    fun toStringOrig(SOME(ORIG(cl))) = " orig {"^toStringSymbList(cl)^"}"
      | toStringOrig(NONE)           = ""
      | toStringOrig(_)              = raise Problem "unsupported annotation" 

				    
    fun toStringTerm(NAME(n))        = toStringSymb(n)
      | toStringTerm(NAMEP(n))       = toStringSymbPM "+" (n)
      | toStringTerm(NAMEM(n))       = toStringSymbPM "-" (n)
      | toStringTerm(VAR(x))         = toStringSymb(x)
      | toStringTerm(ENC(tl,t,c,a))  = "{ "^toStringTermList(tl)^" } : "^
				       toStringTerm(t)^
				       " [at "^toStringSymb(c)^toStringDest(a)^"]"
      | toStringTerm(AENC(tl,t,c,a)) = "{| "^toStringTermList(tl)^" |} : "^
				       toStringTerm(t)^
				       " [at "^toStringSymb(c)^toStringDest(a)^"]"

    and toStringTermList([])         = ""
      | toStringTermList([t])        = toStringTerm(t)
      | toStringTermList(t::tl)      = toStringTerm(t)^","^toStringTermList(tl)


    fun toStringProc(OUT(tl,P))      = "<"^toStringTermList(tl)^">.\n"^
				     toStringProc(P)

      | toStringProc(INP(tl,xl,P))  = "("^toStringTermList(tl)^
                                      "; "^toStringTermList(xl)^").\n"^		    
				      toStringProc(P)

      | toStringProc(DEC(t,tl,xl,t0,c,a,P)) 
                                  = "decrypt "^toStringTerm(t)^" as "^
				    "{ "^ toStringTermList(tl)^
				    "; "^toStringTermList(xl)^
				    " } : "^toStringTerm(t0)^
				    " [at "^toStringSymb(c)^toStringOrig(a)^"]"^
				    " in\n"^toStringProc(P)
      | toStringProc(ADEC(t,tl,xl,t0,c,a,P)) 
	                           = "decrypt "^
				     toStringTerm(t)^" as "^
				     "{| "^toStringTermList(tl)^
				     "; "^toStringTermList(xl)^
				     " |} : "^toStringTerm(t0)^
				     " [at "^toStringSymb(c)^toStringOrig(a)^"] "^
				     "in\n"^toStringProc(P)
      | toStringProc(PAR([]))         = ""
      | toStringProc(PAR([P]))        = toStringProc(P)
      | toStringProc(PAR(P::pl))      = toStringProc(P)^"\n|\n"^toStringProc(PAR(pl))
      | toStringProc(BANG(P))         = "! "^toStringProc(P)
      | toStringProc(NEW(n,P))        = "(new "^toStringTerm(n)^")(\n"^toStringProc(P)^")"
      | toStringProc(ANEW(n,P))       = "(new +- "^toStringTerm(n)^")(\n"^toStringProc(P)^")"
      | toStringProc(NIL)             = "0"			      

      | toStringProc(PAR_I(i,a,P))    = "|_{"^i^"="^toStringNum(a)^"} "^toStringProc(P)
      | toStringProc(PAR_X(i,a,j,P))  = "|_{"^i^"="^toStringNum(a)^"\\"^j^"} "^toStringProc(P)

      | toStringProc(NEW_I(ls,P)) = let fun toStringSingle(il,n) = 
					  "(new_{"^
					  (between 
					       (List.map (fn (i,a) => i^"="^toStringNum(a)^"") il)
					       ",")^
					  "} "^toStringTerm(n)^")\n"					      
				      in
					  String.concat(List.map toStringSingle ls)^
					  "(\n"^toStringProc(P)^")"
				      end

      | toStringProc(ANEW_I(ls,P))  = let fun toStringSingle(il,n) = 
					  "(new_{"^
					  (between 
					       (List.map (fn (i,a) => i^"="^toStringNum(a)^"") il)
					       ",")^
					  "} +- "^toStringTerm(n)^")\n"
				      in
					  String.concat(List.map toStringSingle ls)^
					  "(\n"^toStringProc(P)^")"
				      end
				      
      | toStringProc(_)   = raise Problem "No text syntax for this element element"
in
    fun toString P = toStringProc(P)

    fun toFile file P = let val fh = TextIO.openOut(file)
			    in
				(TextIO.output(fh, toStringProc(P));
				 TextIO.closeOut(fh))
			    end
end	
end