(*****************************************************************************)
(* Data structure and auxiliary functions for Labelled Lysa (LLysa)

                                   Time-stamp: <14-01-2004 Mikael Buchholtz> *)
(*****************************************************************************)


(*****************************************************************************)
(*                                                                           *)
(* Datatype for Labelled LySa is defined below.                              *)
(*                                                                           *)
(* NOTE: as for LySa the variables and names in input, decryption, and       *)
(* restriction are represented as terms (which means that they are also      *)
(* labelled)                                                                 *)
(*                                                                           *)
(*****************************************************************************)


structure LLysa =
struct

type Name     = string
type Var      = string
type CP       = string (* Crypto-points *)
type Lab      = string (* Term Labels *)


datatype Term = NAME  of Name * Lab
	      | NAMEP of Name * Lab
	      | NAMEM of Name * Lab
	      | VAR   of Var  * Lab
	      | ENC   of Term list * Term * CP * CP list * Lab * Lab 
	      | AENC  of Term list * Term * CP * CP list * Lab * Lab 

datatype Proc = OUT   of Term list * Proc
 	      | INP   of Term list * Term list * Proc			 
	      | DEC   of Term *  Term list * Term list * Term *
			 CP * CP list * Lab * Proc	       
	      | ADEC  of Term *  Term list * Term list * Term *
			 CP * CP list * Lab * Proc	       
	      | NEW   of Term * Proc
	      | ANEW  of Term * Proc
	      | BANG  of Proc
	      | PAR   of Proc * Proc
	      | NIL

val CPDY = "BULLET"

exception Problem of string

local 
   open Set
in

(*****************************************************************************)
(* namesProc : LProc -> LName list                                           *)
(* Returns the names in a process                                            *)
(* Labels at names are ignored                                               *)
(*****************************************************************************)

fun namesTerm(NAME(n,_))    = [NAME(n,"")]
  | namesTerm(NAMEP(n,_))   = [NAMEP(n,"")]
  | namesTerm(NAMEM(n,_))   = [NAMEM(n,"")]
  | namesTerm(VAR(_,_))     = []
  | namesTerm(ENC(el,e0,c,cl,ls,l))  = unionList(List.map namesTerm (e0::el))
  | namesTerm(AENC(el,e0,c,cl,ls,l)) = unionList(List.map namesTerm (e0::el))

				      
fun namesProc(OUT(el,p))    = union(unionList(List.map namesTerm el), 
				    namesProc(p))  
  | namesProc(INP(el,xl,p)) = union(unionList(List.map namesTerm el),
				    namesProc(p))
  | namesProc(DEC(e,el,xl,e0,c,cl,ls,p))  
                            = union(unionList (map namesTerm (e::e0::el)),
				    namesProc(p))
  | namesProc(ADEC(e,el,xl,e0,c,cl,ls,p)) 
                            = union(unionList (map namesTerm (e::e0::el)),
				    namesProc(p))
  | namesProc(NEW(NAME(n,_),p)) 
                            = union([NAME(n,"")], namesProc(p))
  | namesProc(NEW(_,_))     = raise Problem "Restriction must restrict a name"
  | namesProc(ANEW(NAME(n,_),p)) 
                            = union([NAMEP(n,""), NAMEM(n,"")], namesProc(p))
  | namesProc(ANEW(_,_))    = raise Problem "Restriction must restrict a name"
  | namesProc(BANG(p))      = namesProc(p)
  | namesProc(PAR(p1,p2))   = union(namesProc(p1), namesProc(p2))
  | namesProc(NIL)          = []



(*****************************************************************************)
(* fnProc: Proc -> Name list                                                 *)
(* Returns the free names in a process                                       *)
(* Labels are ignored and are instead replaced with LDY                      *)
(*****************************************************************************)
 
fun fnTerm(NAME(n,_))         = [NAME(n,"")]
  | fnTerm(NAMEP(n,_))        = [NAMEP(n,"")]
  | fnTerm(NAMEM(n,_))        = [NAMEM(n,"")]
  | fnTerm(VAR(_,_))          = []
  | fnTerm(ENC(el,e0,c,cl,ls,l))  
                              = unionList(List.map fnTerm (e0::el))
  | fnTerm(AENC(el,e0,c,cl,ls,l)) 
                              = unionList(List.map fnTerm (e0::el))

fun fnProc(OUT(el,p))         = union(unionList(List.map fnTerm el),
				      fnProc(p))
  | fnProc(INP(el,xl,p))      = union(unionList(List.map fnTerm el),
				      fnProc(p))				
  | fnProc(DEC(e,el,xl,e0,c,cl,ls,p)) 
                              = union(unionList(List.map fnTerm (e::e0::el)),
				      fnProc(p))
  | fnProc(ADEC(e,el,xl,e0,c,cl,ls,p)) 
                              = union(unionList(List.map fnTerm (e::e0::el)),
				      fnProc(p))
  | fnProc(NEW(NAME(n,_),p))  = difference([NAME(n,"")], fnProc(p))
  | fnProc(NEW(_,_))          = raise Problem "Restriction must restrict a name"
  | fnProc(ANEW(NAME(n,_),p)) = difference([NAMEP(n,""),NAMEM(n,"")], fnProc(p))
  | fnProc(ANEW(_,_))         = raise Problem "Restriction must restrict a name"
  | fnProc(BANG(p))           = fnProc(p)
  | fnProc(PAR(p1,p2))        = union(fnProc(p1),fnProc(p2)) 
  | fnProc(NIL)               = []			    

 

(*****************************************************************************)
(* sencArityProc :  Proc -> int list                                         *)
(*      Finds arities of symmetric encrypted terms in the process.           *)
(*****************************************************************************)

fun sencArityTerm(NAME(_,_))           = []
  | sencArityTerm(NAMEP(_,_))          = []
  | sencArityTerm(NAMEM(_,_))          = []
  | sencArityTerm(VAR(_,_))            = []
  | sencArityTerm(ENC(el,e0,_,_,_,_))  = union([List.length el],
					       unionList(map sencArityTerm (e0::el)))
  | sencArityTerm(AENC(el,e0,_,_,_,_)) = unionList(map sencArityTerm (e0::el))			 


fun sencArityProc(OUT(el,p))           = union(unionList(map sencArityTerm el),
					       sencArityProc(p))
  | sencArityProc(INP(el,_,p))         = union(unionList(map sencArityTerm el),
					       sencArityProc(p))
  | sencArityProc(DEC(e,el,xl,e0,_,_,_,p))
                                       = union([List.length(el) + List.length(xl)],
					 union(unionList(map sencArityTerm (e::e0::el)),
					       sencArityProc(p)))
  | sencArityProc(ADEC(e,el,xl,e0,_,_,_,p)) 
                                       = union(unionList(map sencArityTerm (e::e0::el)),
					      sencArityProc(p))
  | sencArityProc(NEW(_,p))            = sencArityProc(p)
  | sencArityProc(ANEW(_,p))           = sencArityProc(p)
  | sencArityProc(BANG(p))             = sencArityProc(p)
  | sencArityProc(PAR(p1,p2))          = union(sencArityProc(p1),
					       sencArityProc(p2))
  | sencArityProc(NIL)                 = []



(*****************************************************************************)
(* aencArityProc :  Proc -> int list                                         *)
(*      Finds arities of asymmetric encrypted terms in the process.          *)
(*****************************************************************************)

fun aencArityTerm(NAME(_,_))           = []
  | aencArityTerm(NAMEP(_,_))          = []
  | aencArityTerm(NAMEM(_,_))          = []
  | aencArityTerm(VAR(_,_))            = []
  | aencArityTerm(ENC(el,e0,_,_,_,_))  = unionList(map aencArityTerm (e0::el))
  | aencArityTerm(AENC(el,e0,_,_,_,_)) = union([List.length el],
					       unionList(map aencArityTerm (e0::el)))


fun aencArityProc(OUT(el,p))           = union(unionList(map aencArityTerm el),
					       aencArityProc(p))
  | aencArityProc(INP(el,_,p))         = union(unionList(map aencArityTerm el),
					       aencArityProc(p))
  | aencArityProc(DEC(e,el,xl,e0,_,_,_,p))
                                       = union(unionList(map aencArityTerm (e::e0::el)),
					       aencArityProc(p))
  | aencArityProc(ADEC(e,el,xl,e0,_,_,_,p)) 
                                       = union([List.length(el) + List.length(xl)],
                                         union(unionList(map aencArityTerm (e::e0::el)),
					       aencArityProc(p))) 
  | aencArityProc(NEW(_,p))            = aencArityProc(p)
  | aencArityProc(ANEW(_,p))           = aencArityProc(p)
  | aencArityProc(BANG(p))             = aencArityProc(p)
  | aencArityProc(PAR(p1,p2))          = union(aencArityProc(p1),
					       aencArityProc(p2))
  | aencArityProc(NIL)                 = []
 


(*****************************************************************************)
(* kappaArityProc :  Proc -> int list                                        *)
(*      Finds arities of messages send by the process.                       *)
(*****************************************************************************)

fun kappaArityProc(NIL)                  = []
  | kappaArityProc(OUT(el,p))            = union([(List.length el)],
						 kappaArityProc(p))
  | kappaArityProc(INP(el,xl,p))         = union([(List.length el +
						   List.length xl)],
						 kappaArityProc(p))
  | kappaArityProc(PAR(p1,p2))           = union(kappaArityProc(p1),
					       kappaArityProc(p2))
  | kappaArityProc(NEW(_,p))             = kappaArityProc(p)
  | kappaArityProc(ANEW(_,p))            = kappaArityProc(p)
  | kappaArityProc(BANG(p))              = kappaArityProc(p)
  | kappaArityProc(DEC(_,_,_,_,_,_,_,p)) = kappaArityProc(p)
  | kappaArityProc(ADEC(_,_,_,_,_,_,_,p))= kappaArityProc(p)



(*****************************************************************************)
(* cryptoProc: Proc -> CP list                                               *)
(* Returns the  crypto-points in a process                                   *)
(*****************************************************************************)

fun cryptoTerm(NAME(_,_))         = []
  | cryptoTerm(NAMEP(_,_))        = []
  | cryptoTerm(NAMEM(_,_))        = []
  | cryptoTerm(VAR(_,_))          = []
  | cryptoTerm(ENC(el,e0,c,cl,ls,l)) 
                                  = union([c],
				    union(cl,
					  unionList(List.map cryptoTerm (e0::el))))
  | cryptoTerm(AENC(el,e0,c,cl,ls,l)) 
                                  = union([c],
				    union(cl,
					  unionList(List.map cryptoTerm (e0::el))))
				    
fun cryptoProc(NIL)               = []
  | cryptoProc(OUT(el,p))         = union(unionList(List.map cryptoTerm el),
					  cryptoProc(p))
  | cryptoProc(INP(el,xl,p))      = union(unionList(List.map cryptoTerm el),
					  cryptoProc(p))				
  | cryptoProc(PAR(p1,p2))        = union(cryptoProc(p1),cryptoProc(p2))
  | cryptoProc(NEW(_,p))          = cryptoProc(p)
  | cryptoProc(ANEW(_,p))         = cryptoProc(p)
  | cryptoProc(BANG(p))           = cryptoProc(p)
  | cryptoProc(DEC(e,el,xl,e0,c,cl,ls,p)) 
                                  = union([c],
				    union(cl,				    
			            union(unionList(List.map cryptoTerm (e::e0::el)),
					  cryptoProc(p))))
  | cryptoProc(ADEC(e,el,xl,e0,c,cl,ls,p)) 
                                  = union([c],
				    union(cl,				    
			            union(unionList(List.map cryptoTerm (e::e0::el)),
					  cryptoProc(p))))

end

end