(********************************************************************)
(* Data structure for Lysa and Meta Lysa together with auxiliary 
   functions.                                                       

   Adapted from implementation by Hanne Riis Nielson.
                          Time-stamp: <14-01-2004 Mikael Buchholtz> *)
(********************************************************************)

structure Lysa =
struct

type Symbol   = string * string list

type Name     = Symbol
type Var      = Symbol
type CP       = Symbol 

datatype Form = ORIG   of CP list
              | DEST   of CP list

datatype Term = NAME   of Name
	      | NAMEP  of Name
	      | NAMEM  of Name
              | VAR    of Var
              | ENC    of Term list * Term * CP * Form option
              | AENC   of Term list * Term * CP * Form option

datatype Proc = OUT    of Term list * Proc 
              | INP    of Term list * Term list * Proc 
	      | DEC    of Term * Term list * Term list * Term * 
                          CP * Form option * Proc
	      | ADEC   of Term * Term list * Term list * Term * 
                          CP * Form option * Proc
              | NEW    of Term * Proc
	      | ANEW   of Term * Proc
              | BANG   of Proc
	      | PAR    of Proc list           
	      | NIL
           
	      (******************************************************) 
              (* Elements below are conceptually part of Meta Lysa  *)
	      (******************************************************) 
              
              | PAR_I  of string * int * Proc
              | PAR_X  of string * int * string * Proc
              | NEW_I  of ((string * int) list * Term) list * Proc
              | ANEW_I of ((string * int) list * Term) list * Proc

	      (******************************************************) 
              (* Elements below are inserted for "Experiment 1"     *)
	      (******************************************************) 

              | LEAK   of Proc * (string * int) list
              | TEST   of string

val CPDY      = ("\\ell_\\bullet",[])
		
exception Problem of string; 

local
    open Set
in


(********************************************************************)
(* fnProc : Proc -> Term List                                       *)
(* finds free names in a process. The resulting list only contains  *)
(* names i.e. terms on the form  NAME n, NAMEP n, or NAMEM n        *)
(********************************************************************)

fun fnTerm (NAME n)         = [NAME n]
  | fnTerm (NAMEP n)        = [NAMEP n]
  | fnTerm (NAMEM n)        = [NAMEM n]
  | fnTerm (VAR x)          = []
  | fnTerm (ENC(ts,t,_,_))  = unionList((fnTerm t)::(map fnTerm ts))
  | fnTerm (AENC(ts,t,_,_)) = unionList((fnTerm t)::(map fnTerm ts))


fun fnProc (OUT(ts,P))      = unionList((fnProc P)::(map fnTerm ts))
  | fnProc (INP(ts,xs,P))   = unionList((fnProc P)::(map fnTerm ts))
  | fnProc (DEC(t,ts,xs,tk,l,F,P)) 
                            = unionList((fnTerm t)::(fnTerm tk)::
                                     (fnProc P)::(map fnTerm ts))
  | fnProc (ADEC(t,ts,xs,tk,l,F,P)) 
                            = unionList((fnTerm t)::(fnTerm tk)::
                                     (fnProc P)::(map fnTerm ts))
  | fnProc (NEW(NAME(n),P)) = difference ([NAME n], fnProc P)
  | fnProc (NEW(_,_))       = raise Problem "Restriction has to restrict a NAME"
  | fnProc (ANEW(NAME(n),P))= difference ([NAMEP(n), NAMEM(n)], fnProc P)
  | fnProc (ANEW(_,_))      = raise Problem "Restriction has to restrict a NAME"
  | fnProc (BANG P)         = fnProc P
  | fnProc (PAR Ps)         = unionList (map fnProc Ps) 
  | fnProc NIL              = []
  | fnProc (_)              = 
    raise Problem("fnProc is only defined on Lysa, NOT on Meta Lysa")



(********************************************************************)
(* cryptoProc : Proc -> CP list                                     *)
(* cryptoProc find the crypto-points in a Lysa process              *)
(********************************************************************)

fun cryptoForm(SOME(ORIG(cl)))      = unionList(map (fn x => [x]) cl)
  | cryptoForm(SOME(DEST(cl)))      = unionList(map (fn x => [x]) cl)
  | cryptoForm(NONE)                = []

fun cryptoTerm(ENC(el,e0,c,f))      = union(unionList(map cryptoTerm (e0::el)),
					    union([c], 
						  cryptoForm(f)))
  | cryptoTerm(AENC(el,e0,c,f))     = union(unionList(map cryptoTerm (e0::el)),
					    union([c], 
						  cryptoForm(f)))

  | cryptoTerm(_)                   = [] (* no crypto-points in names or vars *)


fun cryptoProc(OUT(el,p))           = union(unionList(map cryptoTerm el),
					    cryptoProc(p))
  | cryptoProc(INP(el,_,p))         = union(unionList(map cryptoTerm el), 
					    cryptoProc(p))
  | cryptoProc(DEC(e,el,_,k,c,f,p)) = union(unionList(map cryptoTerm (e::k::el)),
				      union([c],
				      union(cryptoForm(f), 
					    cryptoProc(p))))
  | cryptoProc(ADEC(e,el,_,k,c,f,p))= union(unionList(map cryptoTerm (e::k::el)),
				      union([c],
				      union(cryptoForm(f), 
					    cryptoProc(p))))
  | cryptoProc(NEW(_,p))            = cryptoProc(p)
  | cryptoProc(ANEW(_,p))           = cryptoProc(p)
  | cryptoProc(BANG(p))             = cryptoProc(p) 
  | cryptoProc(PAR(pl))             = unionList(map cryptoProc pl)
  | cryptoProc(NIL)                 = []

  | cryptoProc (_)                  = 
    raise Problem("cryptoProc is only defined on Lysa, NOT on Meta Lysa")

end


end