(********************************************************************)
(* Functions for transformation of Lysa to Labelled Lysa (LLysa)

                          Time-stamp: <14-01-2004 Mikael Buchholtz> *)
(********************************************************************)

(* Requirements:
   symbols (n,is) will be transformed so that 

      n   should only contain alphanumeric characters or primes (').
          All other characters will be removed with name-clashes as a
          possible consequence.

      is  should only contain strings that are numbers. Negative numbers 
          must be prefixed with -
*)


structure Lysa2LLysa =

struct

local   
    open LLysa	 

    val tlab = ref 0

    val allC = ref [] : string list ref

    fun new() = let val _ = tlab := !tlab + 1
		in "L"^Int.toString(!tlab)
		end

    fun l2llStripString(s) = 
	let fun StripChar(c) = if (Char.isAlphaNum(c) orelse c = #"'") 
			       then Char.toString(c) else ""
	in String.concat(List.map StripChar (String.explode s))
	end


    fun l2llNumbers(s) = 
	let val sl = String.explode s
	in  case sl of 
		([]) => ""			    
	      | (_)  => if List.hd sl = #"-" 
			then "_"^(l2llStripString(String.implode (List.tl sl)))
			else l2llStripString s
	end

    fun l2llSymbol(s,sl) = 
	l2llStripString(s)^"_"^String.concat(map l2llNumbers(sl))
    
    fun l2llSymbol(s,[]) = l2llStripString(s)^"_"
      | l2llSymbol(s,sl) = l2llStripString(s)^String.concat(map (fn x => ("_"^l2llNumbers(x))) sl)
in

fun reset() = (tlab:=0 )


fun l2llCP((c,cl)) = if (c,cl)=Lysa.CPDY then CPDY else l2llSymbol(c,cl)

							
fun l2llOrig(SOME(Lysa.ORIG(ll))) = map l2llCP ll
  | l2llOrig(SOME(Lysa.DEST(ll))) = raise Lysa.Problem "ORIG expected but DEST found."
  | l2llOrig(NONE)                = !allC 


fun l2llDest(SOME(Lysa.ORIG(ll))) = raise Lysa.Problem "DEST expected but ORIG found."
  | l2llDest(SOME(Lysa.DEST(ll))) = map l2llCP ll			    
  | l2llDest(NONE)                = !allC 



(*****************************************************************************)
(* l2llProc : Lysa.Proc -> LLysa.Proc                                        *)
(*                                                                           *)
(* Transforms  LySa to Labelled Lysa and adds unique labels.                 *)
(*****************************************************************************)

fun l2llTerm(Lysa.NAME(n,nl))     = NAME(l2llSymbol(n,nl),new())
  | l2llTerm(Lysa.NAMEP(n,nl))    = NAMEP(l2llSymbol(n,nl),new())
  | l2llTerm(Lysa.NAMEM(n,nl))    = NAMEM(l2llSymbol(n,nl),new())
  | l2llTerm(Lysa.VAR(x,xl))      = VAR(l2llSymbol(x,xl),new())
  | l2llTerm(Lysa.ENC(el,e,c,f))  = ENC((map l2llTerm el),
					l2llTerm(e),
					l2llCP(c),
					l2llDest(f),
					new(),
					new())
  | l2llTerm(Lysa.AENC(el,e,c,f)) = AENC((map l2llTerm el),
					 l2llTerm(e),
					 l2llCP(c),
					 l2llDest(f),
					 new(),
					 new())
local					
					
fun l2llProc'(Lysa.NIL)            = NIL
  | l2llProc'(Lysa.OUT(el,p))      = OUT(map l2llTerm el, l2llProc'(p))
  | l2llProc'(Lysa.INP(el,xl,p))   = INP(map l2llTerm el, map l2llTerm xl, l2llProc'(p))
  | l2llProc'(Lysa.PAR(pl))        = (case pl of
					 []     => NIL
				       | [p]    => l2llProc'(p)
				       | p::pl1 => PAR(l2llProc'(p), 
						      l2llProc'(Lysa.PAR(pl1))))		    
  | l2llProc'(Lysa.NEW(Lysa.NAME(n),p)) 
                                  = NEW(l2llTerm(Lysa.NAME(n)), l2llProc'(p))
  | l2llProc'(Lysa.NEW(_,_))       = raise Problem  "Restriction does not contain a name"

  | l2llProc'(Lysa.ANEW(Lysa.NAME(n),p))
                                  = ANEW(l2llTerm(Lysa.NAME(n)), l2llProc'(p))
  | l2llProc'(Lysa.ANEW(_,_))      = raise Problem  "Restriction does not contain a name"
  | l2llProc'(Lysa.BANG(p))        = BANG(l2llProc'(p))
  | l2llProc'(Lysa.DEC(e,el,xl,e0,c,f,p)) 
                                  = DEC(l2llTerm(e),				     
					List.map l2llTerm el,
					List.map l2llTerm xl,
					l2llTerm(e0),
					l2llCP(c),
					l2llOrig(f),
					new(),
					l2llProc'(p))
  | l2llProc'(Lysa.ADEC(e,el,xl,e0,c,f,p)) 
                                  = ADEC(l2llTerm(e),				     
					 List.map l2llTerm el,
					 List.map l2llTerm xl,
					 l2llTerm(e0),
					 l2llCP(c),
					 l2llOrig(f),
					 new(),
					 l2llProc'(p))
  | l2llProc'(_)                   = raise Lysa.Problem("l2llProc' only defined on LySa"^
						       " - NOT on Meta LySa!")
			

in
fun l2llProc(P) =
    let val _ = (allC := List.map l2llCP (Lysa.CPDY::Lysa.cryptoProc(P)))
    in
	l2llProc'(P)
    end;
end

(*
fun l2sString(s) =
    if s = CPDY then Lysa.CPDY
    else s
*)

end
end