(*****************************************************************************)
(* Functions for transformation of Meta Lysa to Lysa

   Adapted from implementation by Hanne Riis Nielson.
                                   Time-stamp: <14-01-2004 Mikael Buchholtz> *)
(*****************************************************************************)

(******************************************************************************) 
(* Main function:                                                             *)
(*    ml2lProc :   transforms a Meta Lysa process to Lysa process with        *)
(*                 b instances                                                *)
(******************************************************************************)    


structure MLysa2Lysa =

struct

local
    open Lysa

    (*************************************************************************) 
    (* Unfolding transformation for protocols for given b                    *)
    (*                                                                       *)
    (* In subst functions: i is the index and a is the value of the index    *)
    (*                                                                       *) 
    (*  - substIdx:    replaces i with a in a sequence of indices            *)
    (*  - substSymbol: replaces i with a in a symbol                         *)
    (*  - substTerm:   replaces i with a in a term                           *)
    (*  - substForm:   replaces i with a in an annotation                    *)
    (*  - substProc:   replaces i with a in a process                        *)
    (*************************************************************************) 


    (* integer to string with "nice" minus ******)

    fun int2str a                  = if a < 0
                                     then "-"^Int.toString (~a)
                                     else Int.toString a    
in


    (* substitution on a sequence of indices ******)

    fun substIdx (i,a) (i'::is)    = if i=i' 
                                     then (int2str a) :: is 
                                     else i'::substIdx (i,a) is
      | substIdx (i,a) []          = []


    (* substitution on symbols i.e. onName, Var, and CP ******)

    fun substSymbol (i,a) (n,is)   = (n, substIdx (i,a) is)
				     

    (* substitution on crypto-points ******)
    fun substCP (i,a) (n,is)    =  (n, substIdx (i,a) is)


    (* substitution on crypto-points in sets of crypto-points ******)
    (* if the value 0 is in the index then CPDY is added to the set *)

    fun substCPS (i,a) (n,is)   =  if a = 0
                                   then if List.exists (fn i' => i=i') is
					then [(n, substIdx (i,a) is),CPDY]
					else [(n, substIdx (i,a) is)]
				   else [(n, substIdx (i,a) is)]


    (* substitution on annotations ******)

    fun substForm (i,a) (ORIG ls)  = ORIG (Set.unionList (List.map (substCPS (i,a)) ls) )
      | substForm (i,a) (DEST ls)  = DEST (Set.unionList (List.map (substCPS (i,a)) ls) )


    (* substitution on terms ******)

    fun substTerm (i,a) (NAME n)   = NAME (substSymbol (i,a) n)
      | substTerm (i,a) (NAMEP n)  = NAMEP(substSymbol (i,a) n)
      | substTerm (i,a) (NAMEM n)  = NAMEM(substSymbol (i,a) n)
      | substTerm (i,a) (VAR x)    = VAR  (substSymbol (i,a) x)
      | substTerm (i,a) (ENC (ts,t,l,F)) 
                                   = ENC (map (substTerm (i,a)) ts, 
					  substTerm (i,a) t,
					  substCP (i,a) l,
					  Option.map (substForm (i,a)) F)
      | substTerm (i,a) (AENC (ts,t,l,F)) 
                                   = AENC(map (substTerm (i,a)) ts, 
					  substTerm (i,a) t,
					  substCP (i,a) l,
					  Option.map (substForm (i,a)) F)

				      
    (* substitution on processes; *)

    fun substProc (i,a) (OUT(ts,P)) 
                                   = OUT(map (substTerm (i,a)) ts, 
					 substProc (i,a) P)     
      | substProc (i,a) (INP(ts,vs,P)) 
                                   = INP(map (substTerm (i,a)) ts, 
					 map (substTerm (i,a)) vs,
					 substProc (i,a) P) 

      | substProc (i,a) (DEC(t,ts,vs,t',l,F,P)) 
                                   = DEC (substTerm (i,a) t, 
					  map (substTerm (i,a)) ts, 
					  map (substTerm (i,a)) vs, 
					  substTerm (i,a) t',
					  substCP (i,a) l,
					  Option.map (substForm (i,a)) F,
					  substProc (i,a) P)
     
      | substProc (i,a) (ADEC(t,ts,vs,t',l,F,P)) 
                                   = ADEC(substTerm (i,a) t, 
					  map (substTerm (i,a)) ts, 
					  map (substTerm (i,a)) vs, 
					  substTerm (i,a) t',
					  substCP (i,a) l,
					  Option.map (substForm (i,a)) F,
					  substProc (i,a) P)
     

      | substProc (i,a) (NEW(t,P)) = NEW (substTerm (i,a) t, 
					  substProc (i,a) P)
				     
      | substProc (i,a) (ANEW(t,P))= ANEW (substTerm (i,a) t, 
					   substProc (i,a) P)

      | substProc (i,a) (BANG P)   = BANG (substProc (i,a) P)

      | substProc (i,a) (PAR Ps)   = PAR (map (substProc (i,a)) Ps)

      | substProc (i,a)  NIL       = NIL

      | substProc (i,a) (PAR_I(i',a',P)) 
	                            = if i = i' 
				      then PAR_I(i',a',P)
				      else PAR_I(i',a',substProc (i,a) P)

      | substProc (i,a) (PAR_X(i',a',j,P)) 
	                            = if i = i' 
				      then if i=j 
					   then PAR_X(i',a',Int.toString(a),P)
					   else PAR_X(i',a',j,P)					
				      else if i=j 
				           then PAR_X(i',a',Int.toString(a),substProc (i,a) P)
				           else PAR_X(i',a',j,substProc (i,a) P)

      | substProc (i,a) (NEW_I(ias,P)) 
                                    = NEW_I(ias, substProc (i,a) P)

      | substProc (i,a) (ANEW_I(ias,P)) 
                                    = ANEW_I(ias, substProc (i,a) P)
					   
      | substProc (i,a) (LEAK(P,(i',a')::ls)) 
                                    = if i = i'
                                      then if a = a'
					   then LEAK(substProc (i,a) P,ls)
					   else NIL
                                      else (case substProc (i,a) (LEAK(P,ls)) of
						LEAK(P',ls') => LEAK(P',(i',a')::ls') |
						P' => (P') )


      | substProc (i,a) (LEAK(P,[])) = LEAK(P,[])

      | substProc (i,a) (TEST s)     = TEST s 


    (* ml2lProc will unfold the process b times where specified *)

    fun ml2lProc b (OUT(ts,P))    = OUT(ts,ml2lProc b P)
      | ml2lProc b (INP(ts,vs,P)) = INP(ts,vs,ml2lProc b P)
      | ml2lProc b (DEC(e,xl,el,e0,c,F,P))
		                  = DEC(e,xl,el,e0,c,F,ml2lProc b P)
      | ml2lProc b (ADEC(e,xl,el,e0,c,F,P))
		                  = ADEC(e,xl,el,e0,c,F,ml2lProc b P)
				    
      | ml2lProc b (NEW(n,P))     = NEW(n, ml2lProc b P)
      | ml2lProc b (ANEW(n,P))    = ANEW(n, ml2lProc b P)
      | ml2lProc b (BANG(P))      = BANG(ml2lProc b P)
      | ml2lProc b (PAR Ps)       = PAR (map (ml2lProc b) Ps)
      | ml2lProc b (NIL)          = NIL

      | ml2lProc b (PAR_I(i,a,P)) = 
	let fun unfold (i,a) P = if a <= b
				 then (substProc (i,a) P)::(unfold (i,a+1) P)
				 else []
	in
	    ml2lProc b (PAR(unfold (i,a) P))
	end

      | ml2lProc b (PAR_X(i,a,j,P)) = 
	let fun unfold (i,a) P = if a <= b
				 then if Int.toString(a) = j 
				      then unfold (i,a+1) P
				      else (substProc (i,a) P)::(unfold (i,a+1) P)
				 else []
	in
	    ml2lProc b (PAR(unfold (i,a) P))
	end
	                          
    
      | ml2lProc b (NEW_I ((ias,NAME((n,nis)))::ts,P)) =
	let	    
	    fun unfoldIdx (i,a) mis = if a <= b 
				      then (substIdx(i,a) mis)
					   ::(unfoldIdx (i,a+1) mis)
				      else []
				       
	    fun genIdx ((i,a)::iss) mis = genIdx iss (List.concat
							  (List.map (unfoldIdx (i,a)) mis))
	      | genIdx ([])         mis = mis
					  
	in
	    List.foldr 
		(fn(is,P') => NEW(NAME(n,is), P')) 
		(ml2lProc b (NEW_I ((ts,P))))
		(genIdx ias [nis])
	end

      | ml2lProc b (NEW_I ((ias,_)::ts,P)) 
	                          = raise Problem "Restrictions must restrict a name"

      | ml2lProc b (NEW_I([],P))  = ml2lProc b P


      | ml2lProc b (ANEW_I ((ias,NAME((n,nis)))::ts,P)) =
	let	    
	    fun unfoldIdx (i,a) mis = if a <= b 
				      then (substIdx(i,a) mis)
					   ::(unfoldIdx (i,a+1) mis)
				      else []
				       
	    fun genIdx ((i,a)::iss) mis = genIdx iss (List.concat
							  (List.map (unfoldIdx (i,a)) mis))
	      | genIdx ([])         mis = mis
					  
	in
	    List.foldr 
		(fn(is,P') => ANEW(NAME(n,is), P')) 
		(ml2lProc b (ANEW_I ((ts,P))))
		(genIdx ias [nis])
	end

      | ml2lProc b (ANEW_I ((ias,_)::ts,P)) 
	                          = raise Problem "Restrictions must restrict a name"

      | ml2lProc b (ANEW_I([],P)) = ml2lProc b P

      | ml2lProc b (LEAK(P,ls))   = ml2lProc b P
      
      | ml2lProc b (TEST(_))      = raise Problem "ml2lProc not defined for TEST"

end
end